from pathlib import Path
from typing import TYPE_CHECKING, Awaitable, Callable
from uuid import uuid4

try:
    import yappi
except ImportError:  # pragma: no cover
    yappi = None

if TYPE_CHECKING:
    from fastapi import FastAPI, Request, Response


def set_yappi_profiler(app: "FastAPI") -> None:
    if yappi is None:
        msg = (
            "Profiling is not available without yappi installed. "
            "Either install it with `pip install yak-server[profiling]` or disable profiling."
        )
        raise NotImplementedError(msg)

    @app.middleware("http")
    async def profile_process_time(
        request: "Request",
        call_next: Callable[["Request"], Awaitable["Response"]],
    ) -> "Response":
        yappi.clear_stats()
        yappi.set_clock_type("cpu")
        yappi.start()

        response = await call_next(request)

        yappi.stop()

        folder_location = Path(__file__).parents[2] / "profiling"
        folder_location.mkdir(parents=True, exist_ok=True)

        profiling_log_id = uuid4()

        yappi.convert2pstats(yappi.get_func_stats()).dump_stats(
            folder_location / f"{profiling_log_id}.pstats"
        )

        response.headers["profiling-log-id"] = str(profiling_log_id)

        return response



import logging

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.config import Config
from strawberry.fastapi import GraphQLRouter

from .helpers.logging import setup_logging
from .helpers.profiling import set_yappi_profiler
from .v1.helpers.errors import set_exception_handler
from .v1.routers import bets as bets_router
from .v1.routers import binary_bets as binary_bets_router
from .v1.routers import groups as groups_router
from .v1.routers import phases as phases_router
from .v1.routers import results as results_router
from .v1.routers import rules as rules_router
from .v1.routers import score_bets as score_bets_router
from .v1.routers import teams as teams_router
from .v1.routers import users as users_router
from .v2 import get_schema
from .v2.context import get_context

logger = logging.getLogger(__name__)

GLOBAL_ENDPOINT = "api"
VERSION1 = "v1"
VERSION2 = "v2"

__version__ = "0.40.0"


def create_app() -> FastAPI:
    # Initialize fastapi application
    config = Config(".env")

    app = FastAPI(
        debug=config("DEBUG", cast=bool, default=False),
        docs_url=f"/{GLOBAL_ENDPOINT}/docs",
        redoc_url=f"/{GLOBAL_ENDPOINT}/redoc",
        openapi_url=f"/{GLOBAL_ENDPOINT}/openapi.json",
    )

    # Include all routers
    v1_prefix = f"/{GLOBAL_ENDPOINT}/{VERSION1}"

    app.include_router(bets_router.router, prefix=v1_prefix)
    app.include_router(binary_bets_router.router, prefix=v1_prefix)
    app.include_router(groups_router.router, prefix=v1_prefix)
    app.include_router(phases_router.router, prefix=v1_prefix)
    app.include_router(results_router.router, prefix=v1_prefix)
    app.include_router(rules_router.router, prefix=v1_prefix)
    app.include_router(score_bets_router.router, prefix=v1_prefix)
    app.include_router(teams_router.router, prefix=v1_prefix)
    app.include_router(users_router.router, prefix=v1_prefix)

    # Set error handler
    set_exception_handler(app)

    # Register graphql endpoint
    graphql_app = GraphQLRouter(
        get_schema(debug=app.debug),
        graphql_ide="apollo-sandbox" if app.debug is True else None,
        context_getter=get_context,
    )

    app.include_router(graphql_app, prefix=f"/{GLOBAL_ENDPOINT}/{VERSION2}", tags=["graphql"])

    # Set CORS
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Declare logger configuration for yak server
    setup_logging(debug=app.debug)

    # Set yappi profiler
    profiling = config("PROFILING", cast=bool, default=False)

    if app.debug and profiling:
        set_yappi_profiler(app)

    return app



"""Adding english description for Phase, Group and Team

Revision ID: 314114c0b886
Revises: 045a9142794e
Create Date: 2023-05-26 23:46:38.127268

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "314114c0b886"
down_revision = "045a9142794e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("group", sa.Column("description_en", sa.String(length=100), nullable=False))
    op.create_unique_constraint(None, "group", ["description_en"])
    op.add_column("phase", sa.Column("description_en", sa.String(length=100), nullable=False))
    op.add_column("team", sa.Column("description_en", sa.String(length=100), nullable=False))
    op.create_unique_constraint(None, "team", ["description_en"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "team", type_="unique")
    op.drop_column("team", "description_en")
    op.drop_column("phase", "description_en")
    op.drop_constraint(None, "group", type_="unique")
    op.drop_column("group", "description_en")
    # ### end Alembic commands ###



from functools import lru_cache

import pymysql
from pydantic_settings import BaseSettings, SettingsConfigDict
from sqlalchemy import create_engine
from sqlalchemy.orm import declarative_base, sessionmaker


class MySQLSettings(BaseSettings):
    user_name: str = ""
    password: str = ""
    port: int = 3306
    db: str = ""

    model_config = SettingsConfigDict(
        env_file=".env.mysql",
        env_file_encoding="utf-8",
        env_prefix="mysql_",
    )


@lru_cache(maxsize=None)
def get_mysql_settings() -> MySQLSettings:
    return MySQLSettings()


def compute_database_uri(
    mysql_client: str,
    mysql_user_name: str,
    mysql_password: str,
    mysql_port: int,
    mysql_db: str,
) -> str:
    return f"mysql+{mysql_client}://{mysql_user_name}:{mysql_password}@127.0.0.1:{mysql_port}/{mysql_db}"


mysql_settings = get_mysql_settings()


SQLALCHEMY_DATABASE_URL = compute_database_uri(
    pymysql.__name__,
    mysql_settings.user_name,
    mysql_settings.password,
    mysql_settings.port,
    mysql_settings.db,
)

engine = create_engine(SQLALCHEMY_DATABASE_URL, pool_recycle=7200, pool_pre_ping=True)
SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=engine)

Base = declarative_base()



import json
import secrets
from enum import Enum
from pathlib import Path
from typing import Dict
from uuid import UUID

import pendulum
import typer

from yak_server.database import MySQLSettings
from yak_server.helpers.rules import RULE_MAPPING
from yak_server.helpers.settings import Rules


class YesOrNo(str, Enum):
    y = "y"
    n = "n"


class RuleNotDefinedError(Exception):
    def __init__(self, rule_id: str) -> None:
        super().__init__(f"Rule not defined: {rule_id}")


def write_env_file(env: dict, filename: str) -> None:
    Path(filename).write_text(
        "".join([f"{env_var}={env_value}\n" for env_var, env_value in env.items()])
    )


class EnvBuilder:
    def __init__(self) -> None:
        self.env = {}
        self.env_mysql = {}

        debug = typer.prompt("DEBUG (y/n)", type=YesOrNo)

        self.debug = debug == YesOrNo.y
        self.env["DEBUG"] = 1 if self.debug else 0

    def setup_profiling(self) -> None:
        if self.debug:
            profiling = typer.prompt("PROFILING (y/n)", type=YesOrNo)

            self.env["PROFILING"] = 1 if profiling == YesOrNo.y else 0

    def setup_mysql_env(self) -> None:
        user_name = typer.prompt("MYSQL USER NAME")
        password = typer.prompt("MYSQL PASSWORD", hide_input=True)
        port = typer.prompt("MYSQL PORT", type=int, default=3306)
        db = typer.prompt("MYSQL DB")

        # try to instance pydantic model to check if settings are ok
        MySQLSettings(user_name=user_name, password=password, port=port, db=db)

        self.env_mysql["MYSQL_USER_NAME"] = user_name
        self.env_mysql["MYSQL_PASSWORD"] = password
        self.env_mysql["MYSQL_PORT"] = port
        self.env_mysql["MYSQL_DB"] = db

    def setup_jwt(self) -> None:
        jwt_expiration_time = typer.prompt("JWT_EXPIRATION_TIME")

        self.env["JWT_EXPIRATION_TIME"] = jwt_expiration_time
        self.env["JWT_SECRET_KEY"] = secrets.token_hex(128)

    def choose_competition(self) -> None:
        # Select competition to load associated rules
        path = Path(__file__).parents[1] / "data"

        available_competitions = sorted(
            (competition.stem for competition in path.glob("*")), key=lambda x: x.split("_")[-1]
        )

        for index, competition in enumerate(available_competitions, 1):
            print(f"{index} - {competition}")

        competition_choice = typer.prompt("Choose your competition", type=int)

        competition = available_competitions[competition_choice - 1]
        self.env["COMPETITION"] = competition

        data_folder = path / competition
        self.env["DATA_FOLDER"] = data_folder

        # Load rules in environment
        rules_list: Dict[str, dict] = {}

        for rule_file in Path(data_folder, "rules").glob("*.json"):
            rule_id = UUID(rule_file.stem)

            if rule_id not in RULE_MAPPING:
                raise RuleNotDefinedError(rule_id)

            rule_name = RULE_MAPPING[rule_id].attribute

            rules_list[rule_name] = json.loads(rule_file.read_text())

        rules = Rules.model_validate(rules_list)
        self.env["RULES"] = rules.model_dump_json(exclude_unset=True)

        # Load lock datetime
        common_settings = json.loads((data_folder / "common.json").read_text())

        self.env["LOCK_DATETIME"] = pendulum.parse(
            common_settings["lock_datetime"]
        ).to_iso8601_string()
        self.env["OFFICIAL_RESULTS_URL"] = common_settings["official_results_url"]

    def write(self) -> None:
        write_env_file(self.env, ".env")
        write_env_file(self.env_mysql, ".env.mysql")


def init_env() -> None:
    env_builder = EnvBuilder()
    env_builder.setup_profiling()
    env_builder.setup_mysql_env()
    env_builder.setup_jwt()
    env_builder.choose_competition()

    env_builder.write()



import sys
from typing import Optional

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated


from fastapi import Depends
from sqlalchemy.orm import Session
from strawberry.fastapi import BaseContext

from yak_server.database.models import UserModel
from yak_server.helpers.database import get_db
from yak_server.helpers.settings import Settings, get_settings


class YakContext(BaseContext):
    db: Session
    settings: Settings
    user: Optional[UserModel]

    def __init__(self, db: Session, settings: Settings, user: Optional[UserModel] = None) -> None:
        super().__init__()
        self.db = db
        self.settings = settings
        self.user = user


def get_context(
    db: Annotated[Session, Depends(get_db)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> YakContext:
    return YakContext(db, settings)



import logging
from typing import Optional
from uuid import UUID

import pendulum
import strawberry
from strawberry.types import Info

from yak_server.database.models import (
    BinaryBetModel,
    MatchModel,
    MatchReferenceModel,
    ScoreBetModel,
    UserModel,
)
from yak_server.helpers.authentication import encode_bearer_token
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.group_position import create_group_position, set_recomputation_flag
from yak_server.helpers.logging import (
    logged_in_successfully,
    modify_binary_bet_successfully,
    modify_score_bet_successfully,
    signed_up_successfully,
)

from .bearer_authentication import (
    is_admin_authenticated,
    is_authenticated,
)
from .context import YakContext
from .result import (
    BinaryBetNotFoundForUpdate,
    InvalidCredentials,
    LockedBinaryBetError,
    LockedScoreBetError,
    LoginResult,
    ModifyBinaryBetResult,
    ModifyScoreBetResult,
    ModifyUserResult,
    NewScoreNegative,
    ScoreBetNotFoundForUpdate,
    SignupResult,
    UserNameAlreadyExists,
    UserNotFound,
    UserWithoutSensitiveInfo,
)
from .schema import (
    BinaryBet,
    ScoreBet,
    UserWithToken,
)

logger = logging.getLogger(__name__)


@strawberry.type
class Mutation:
    @strawberry.mutation
    def signup_result(
        self,
        user_name: str,
        password: str,
        first_name: str,
        last_name: str,
        info: Info[YakContext, None],
    ) -> SignupResult:
        db = info.context.db
        settings = info.context.settings

        # Check existing user in db
        existing_user = db.query(UserModel).filter_by(name=user_name).first()
        if existing_user:
            return UserNameAlreadyExists(user_name=user_name)

        # Initialize user and integrate in db
        user = UserModel(
            name=user_name,
            first_name=first_name,
            last_name=last_name,
            password=password,
        )
        db.add(user)
        db.flush()

        # Initialize matches and bets and integrate in db
        for match_reference in db.query(MatchReferenceModel).all():
            match = MatchModel(
                team1_id=match_reference.team1_id,
                team2_id=match_reference.team2_id,
                index=match_reference.index,
                group_id=match_reference.group_id,
            )
            db.add(match)
            db.flush()

            db.add(match_reference.bet_type_from_match.value(user_id=user.id, match_id=match.id))
            db.flush()

        # Create group position records
        db.add_all(create_group_position(db.query(ScoreBetModel).filter_by(user_id=user.id)))
        db.commit()

        token = encode_bearer_token(
            sub=user.id,
            expiration_time=pendulum.duration(seconds=settings.jwt_expiration_time),
            secret_key=settings.jwt_secret_key,
        )

        logger.info(signed_up_successfully(user.name))

        return UserWithToken.from_instance(
            user,
            db=db,
            lock_datetime=settings.lock_datetime,
            token=token,
        )

    @strawberry.mutation
    def login_result(
        self,
        user_name: str,
        password: str,
        info: Info[YakContext, None],
    ) -> LoginResult:
        db = info.context.db
        settings = info.context.settings

        user = UserModel.authenticate(db=db, name=user_name, password=password)

        if not user:
            return InvalidCredentials()

        token = encode_bearer_token(
            sub=user.id,
            expiration_time=pendulum.duration(seconds=settings.jwt_expiration_time),
            secret_key=settings.jwt_secret_key,
        )

        logger.info(logged_in_successfully(user.name))

        return UserWithToken.from_instance(
            user,
            db=db,
            lock_datetime=settings.lock_datetime,
            token=token,
        )

    @strawberry.mutation
    @is_authenticated
    def modify_binary_bet_result(
        self,
        id: UUID,
        is_one_won: Optional[bool],  # noqa: FBT001
        info: Info[YakContext, None],
    ) -> ModifyBinaryBetResult:
        db = info.context.db
        user = info.context.user
        settings = info.context.settings

        bet = db.query(BinaryBetModel).filter_by(user_id=user.id, id=id).first()

        if is_locked(user.name, settings.lock_datetime):
            return LockedBinaryBetError()

        if not bet:
            return BinaryBetNotFoundForUpdate()

        logger.info(modify_binary_bet_successfully(user.name, bet, new_is_one_won=is_one_won))

        bet.is_one_won = is_one_won
        db.commit()

        return BinaryBet.from_instance(bet, db=db, lock_datetime=settings.lock_datetime)

    @strawberry.mutation
    @is_authenticated
    def modify_score_bet_result(
        self,
        id: UUID,
        score1: Optional[int],
        score2: Optional[int],
        info: Info[YakContext, None],
    ) -> ModifyScoreBetResult:
        db = info.context.db
        user = info.context.user
        settings = info.context.settings

        bet = db.query(ScoreBetModel).filter_by(user_id=user.id, id=id).first()

        if is_locked(user.name, settings.lock_datetime):
            return LockedScoreBetError()

        if not bet:
            return ScoreBetNotFoundForUpdate()

        if score1 is not None and score1 < 0:
            return NewScoreNegative(variable_name="$score1", score=score1)

        if score2 is not None and score2 < 0:
            return NewScoreNegative(variable_name="$score2", score=score2)

        if score1 == bet.score1 and score2 == bet.score2:
            return ScoreBet.from_instance(bet, db=db, lock_datetime=settings.lock_datetime)

        set_recomputation_flag(db, bet.match.team1_id, user.id)
        set_recomputation_flag(db, bet.match.team2_id, user.id)

        logger.info(modify_score_bet_successfully(user.name, bet, score1, score2))

        bet.score1 = score1
        bet.score2 = score2
        db.commit()

        return ScoreBet.from_instance(bet, db=db, lock_datetime=settings.lock_datetime)

    @strawberry.mutation
    @is_authenticated
    @is_admin_authenticated
    def modify_user_result(
        self,
        id: UUID,
        password: str,
        info: Info[YakContext, None],
    ) -> ModifyUserResult:
        db = info.context.db

        user = db.query(UserModel).filter_by(id=id).first()

        if not user:
            return UserNotFound(id=id)

        user.change_password(password)
        db.commit()

        return UserWithoutSensitiveInfo.from_instance(user)



import logging
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from yak_server.database.models import BinaryBetModel, ScoreBetModel


def modify_score_bet_successfully(
    user_name: str,
    original_bet: "ScoreBetModel",
    new_score1: Optional[int],
    new_score2: Optional[int],
) -> str:
    return (
        f"{user_name} modify "
        f"{original_bet.match.team1.description_en}-{original_bet.match.team2.description_en} "
        f"in {original_bet.match.group.description_en} "
        f"from {original_bet.score1}-{original_bet.score2} "
        f"to {new_score1}-{new_score2}"
    )


def modify_binary_bet_successfully(
    user_name: str,
    original_bet: "BinaryBetModel",
    *,
    new_is_one_won: Optional[bool],
) -> str:
    return (
        f"{user_name} modify "
        f"{original_bet.match.team1.description_en if original_bet.match.team1 else None}"
        f"-{original_bet.match.team2.description_en if original_bet.match.team2 else None} "
        f"in {original_bet.match.group.description_en} "
        f"from {original_bet.is_one_won}-"
        f"{not original_bet.is_one_won if isinstance(original_bet.is_one_won, bool) else None} "
        f"to {new_is_one_won}-{not new_is_one_won if isinstance(new_is_one_won, bool) else None}"
    )


def signed_up_successfully(user_name: str) -> str:
    return f"{user_name} signed up successfully"


def logged_in_successfully(user_name: str) -> str:
    return f"{user_name} logged in successfully"


def modify_password_successfully(user_name: str) -> str:
    return f"admin user modify {user_name} password"


def setup_logging(*, debug: bool) -> None:
    logging.basicConfig(
        filename="yak.log",
        level=logging.DEBUG if debug else logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )



import warnings
from dataclasses import dataclass
from typing import List, Optional, Tuple

import httpx
from sqlalchemy import and_

from yak_server.database import SessionLocal
from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    MatchModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)
from yak_server.helpers.settings import get_settings

try:
    import bs4
except ImportError:  # pragma: no cover
    bs4 = None  # type: ignore[assignment]

try:
    import lxml  # noqa: S410
except ImportError:  # pragma: no cover
    lxml = None  # type: ignore[assignment]


def parse_score(content: str) -> Tuple[int, int]:
    try:
        scores = next(next(content.children).children)
    except AttributeError:
        return None, None

    try:
        score1_str, score2_str = scores.split("\u2013")
    except ValueError:  # pragma: no cover
        return None, None

    try:
        score1 = int(score1_str)
    except ValueError:  # pragma: no cover
        warnings.warn(f"Couldn't parse score1: '{score1_str}'", stacklevel=2)
        score1 = None

    try:
        score2 = int(score2_str)
    except ValueError:  # pragma: no cover
        warnings.warn(f"Couldn't parse score2: '{score2_str}'", stacklevel=2)
        score2 = None

    return score1, score2


def parse_penalty(content: str) -> bool:
    penalties_tag = content.find("a", string="Penalties")
    score_content = penalties_tag.next.parent.find_next("th").string

    score1, score2 = score_content.split("\u2013")

    return int(score1) > int(score2)


def is_penalties(content: str) -> bool:
    return bool(content.find("a", string="Penalties"))


class SyncOfficialResultsNotAvailableError(Exception):
    def __init__(self) -> None:
        super().__init__(
            "Synchronize official results is not available without beautifulsoup4 installed. "
            "To enable it, please run: pip install yak-server[sync]"
        )


@dataclass
class GroupContainer:
    model: GroupModel
    content: str


@dataclass
class Group:
    index: int


@dataclass
class Team:
    score: int
    description: str
    won: Optional[bool] = None


@dataclass
class Match:
    index: int
    group: Group
    team1: Team
    team2: Team


def extract_matches_from_html(groups: List[GroupContainer]) -> List[Match]:
    matches = []

    for index, group in enumerate(groups, start=1):
        try:
            next_group = groups[index + 1]
        except IndexError:
            next_group = None

        previous_matches_html = (
            frozenset(next_group.content.find_all_previous("div", class_="footballbox"))
            if next_group is not None
            else None
        )

        next_matches_html = group.content.find_all_next("div", class_="footballbox")

        matches_html = [
            match
            for match in next_matches_html
            if previous_matches_html is None or match in previous_matches_html
        ]

        for match_index, match_html in enumerate(matches_html, start=1):
            score1, score2 = parse_score(match_html.find("th", class_="fscore"))

            match = Match(
                index=match_index,
                group=Group(index=group.model.index),
                team1=Team(
                    description=(
                        match_html.find("th", class_="fhome").get_text().replace("\xa0", "")
                    ),
                    score=score1,
                ),
                team2=Team(
                    description=(
                        match_html.find("th", class_="faway").get_text().replace("\xa0", "")
                    ),
                    score=score2,
                ),
            )

            if is_penalties(match_html):
                is_one_won = parse_penalty(match_html)

                match.team1.won = is_one_won
                match.team2.won = not is_one_won

            matches.append(match)

    return matches


def synchronize_official_results() -> None:
    if bs4 is None or lxml is None:
        raise SyncOfficialResultsNotAvailableError

    official_results_url = get_settings().official_results_url

    response = httpx.get(str(official_results_url))

    soup = bs4.BeautifulSoup(response.text, "lxml")

    db = SessionLocal()

    groups = [
        GroupContainer(model=group, content=content.parent.parent)
        for group in db.query(GroupModel).order_by(GroupModel.index)
        for content in soup.find("span", string=group.description_en, class_="mw-headline")
    ]

    matches = extract_matches_from_html(groups)

    admin = db.query(UserModel).filter_by(name="admin").first()

    for match in matches:
        score_bet = (
            admin.score_bets.join(ScoreBetModel.match)
            .join(MatchModel.group)
            .filter(and_(GroupModel.index == match.group.index, MatchModel.index == match.index))
            .first()
        )

        if score_bet is not None:
            score_bet.score1 = match.team1.score
            score_bet.score2 = match.team2.score

        binary_bet = (
            admin.binary_bets.join(BinaryBetModel.match)
            .join(MatchModel.group)
            .join(GroupModel.phase)
            .filter(and_(GroupModel.index == match.group.index, MatchModel.index == match.index))
            .first()
        )

        if binary_bet is not None:
            team1 = db.query(TeamModel).filter_by(description_en=match.team1.description).first()
            team2 = db.query(TeamModel).filter_by(description_en=match.team2.description).first()

            if team1 is not None and team2 is not None:
                binary_bet.match.team1_id = team1.id
                binary_bet.match.team2_id = team2.id

                if match.team1.won is not None:
                    binary_bet.is_one_won = match.team1.won
                else:
                    binary_bet.is_one_won = match.team1.score > match.team2.score

    db.commit()



import sys
from typing import List, Optional
from uuid import UUID

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

from fastapi import APIRouter, Depends
from sqlalchemy import func
from sqlalchemy.orm import Session

from yak_server.database.models import UserModel
from yak_server.helpers.database import get_db
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.errors import NoResultsForAdminUser
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.results import UserResult

router = APIRouter(tags=["results"])


@router.get("/score_board")
def retrieve_score_board(
    _: Annotated[UserModel, Depends(get_current_user)],
    db: Annotated[Session, Depends(get_db)],
) -> GenericOut[List[UserResult]]:
    return GenericOut(
        result=[
            UserResult.from_instance(user, rank=rank)
            for rank, user in enumerate(
                db.query(UserModel)
                .order_by(UserModel.points.desc())
                .where(UserModel.name != "admin"),
                1,
            )
        ],
    )


def compute_rank(db: Session, user_id: UUID) -> Optional[int]:
    subq = (
        db.query(
            UserModel,
            func.row_number().over(order_by=UserModel.points.desc()).label("rownum"),
        )
        .filter(UserModel.name != "admin")
        .subquery()
    )

    rank = db.query(subq.c.rownum).filter(subq.c.id == user_id).first()

    if not rank:
        return None

    return rank[0]


@router.get("/results")
def retrieve_user_results(
    user: Annotated[UserModel, Depends(get_current_user)],
    db: Annotated[Session, Depends(get_db)],
) -> GenericOut[UserResult]:
    rank = compute_rank(db, user.id)

    if rank is None:
        raise NoResultsForAdminUser

    user_result = UserResult.from_instance(
        user,
        rank=rank,
    )

    return GenericOut(result=user_result)



import sys

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

from fastapi import APIRouter, Depends
from sqlalchemy.orm import Session

from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    MatchModel,
    PhaseModel,
    ScoreBetModel,
    UserModel,
)
from yak_server.database.query import (
    bets_from_group_code,
    bets_from_phase_code,
)
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.database import get_db
from yak_server.helpers.group_position import get_group_rank_with_code
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang
from yak_server.helpers.settings import Settings, get_settings
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.errors import GroupNotFound, PhaseNotFound
from yak_server.v1.models.bets import (
    AllBetsResponse,
    BetsByGroupCodeResponse,
    BetsByPhaseCodeResponse,
    GroupRankResponse,
)
from yak_server.v1.models.binary_bets import BinaryBetOut, BinaryBetWithGroupIdOut
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.group_rank import GroupPositionOut
from yak_server.v1.models.groups import GroupOut, GroupWithPhaseIdOut
from yak_server.v1.models.phases import PhaseOut
from yak_server.v1.models.score_bets import ScoreBetOut, ScoreBetWithGroupIdOut

router = APIRouter(prefix="/bets", tags=["bets"])


@router.get("/")
def retrieve_all_bets(
    user: Annotated[UserModel, Depends(get_current_user)],
    db: Annotated[Session, Depends(get_db)],
    settings: Annotated[Settings, Depends(get_settings)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[AllBetsResponse]:
    binary_bets = (
        user.binary_bets.join(BinaryBetModel.match)
        .join(MatchModel.group)
        .order_by(GroupModel.index, MatchModel.index)
    )

    score_bets = (
        user.score_bets.join(ScoreBetModel.match)
        .join(MatchModel.group)
        .order_by(GroupModel.index, MatchModel.index)
    )

    groups = db.query(GroupModel).order_by(GroupModel.index)
    phases = db.query(PhaseModel).order_by(PhaseModel.index)

    return GenericOut(
        result=AllBetsResponse(
            phases=[PhaseOut.from_instance(phase, lang=lang) for phase in phases],
            groups=[GroupWithPhaseIdOut.from_instance(group, lang=lang) for group in groups],
            score_bets=[
                ScoreBetWithGroupIdOut.from_instance(
                    score_bet,
                    locked=is_locked(user.name, settings.lock_datetime),
                    lang=lang,
                )
                for score_bet in score_bets
            ],
            binary_bets=[
                BinaryBetWithGroupIdOut.from_instance(
                    binary_bet,
                    locked=is_locked(user.name, settings.lock_datetime),
                    lang=lang,
                )
                for binary_bet in binary_bets
            ],
        ),
    )


@router.get("/phases/{phase_code}")
def retrieve_bets_by_phase_code(
    phase_code: str,
    user: Annotated[UserModel, Depends(get_current_user)],
    db: Annotated[Session, Depends(get_db)],
    settings: Annotated[Settings, Depends(get_settings)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[BetsByPhaseCodeResponse]:
    phase, groups, score_bets, binary_bets = bets_from_phase_code(db, user, phase_code)

    if not phase:
        raise PhaseNotFound(phase_code)

    return GenericOut(
        result=BetsByPhaseCodeResponse(
            phase=PhaseOut.from_instance(phase, lang=lang),
            groups=[GroupOut.from_instance(group, lang=lang) for group in groups],
            score_bets=[
                ScoreBetWithGroupIdOut.from_instance(
                    score_bet,
                    locked=is_locked(user.name, settings.lock_datetime),
                    lang=lang,
                )
                for score_bet in score_bets
            ],
            binary_bets=[
                BinaryBetWithGroupIdOut.from_instance(
                    binary_bet,
                    locked=is_locked(user.name, settings.lock_datetime),
                    lang=lang,
                )
                for binary_bet in binary_bets
            ],
        ),
    )


@router.get("/groups/{group_code}")
def retrieve_bets_by_group_code(
    group_code: str,
    user: Annotated[UserModel, Depends(get_current_user)],
    db: Annotated[Session, Depends(get_db)],
    settings: Annotated[Settings, Depends(get_settings)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[BetsByGroupCodeResponse]:
    group, score_bets, binary_bets = bets_from_group_code(db, user, group_code)

    if not group:
        raise GroupNotFound(group_code)

    return GenericOut(
        result=BetsByGroupCodeResponse(
            phase=PhaseOut.from_instance(group.phase, lang=lang),
            group=GroupOut.from_instance(group, lang=lang),
            score_bets=[
                ScoreBetOut.from_instance(
                    score_bet,
                    locked=is_locked(user.name, settings.lock_datetime),
                    lang=lang,
                )
                for score_bet in score_bets
            ],
            binary_bets=[
                BinaryBetOut.from_instance(
                    binary_bet,
                    locked=is_locked(user.name, settings.lock_datetime),
                    lang=lang,
                )
                for binary_bet in binary_bets
            ],
        ),
    )


@router.get("/groups/rank/{group_code}")
def retrieve_group_rank_by_code(
    group_code: str,
    user: Annotated[UserModel, Depends(get_current_user)],
    db: Annotated[Session, Depends(get_db)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[GroupRankResponse]:
    group = db.query(GroupModel).filter_by(code=group_code).first()

    if not group:
        raise GroupNotFound(group_code)

    group_rank = get_group_rank_with_code(db, user, group.id)

    return GenericOut(
        result=GroupRankResponse(
            phase=PhaseOut.from_instance(group.phase, lang=lang),
            group=GroupOut.from_instance(group, lang=lang),
            group_rank=[
                GroupPositionOut.from_instance(group_position, lang=lang)
                for group_position in group_rank
            ],
        ),
    )



from typing import TYPE_CHECKING, List

from pydantic import UUID4, BaseModel

from yak_server.helpers.language import Lang, get_language_description

from .phases import PhaseOut

if TYPE_CHECKING:
    from yak_server.database.models import GroupModel


class GroupIn(BaseModel):
    id: UUID4


class GroupOut(BaseModel):
    id: UUID4
    code: str
    description: str

    @classmethod
    def from_instance(cls, group: "GroupModel", *, lang: Lang) -> "GroupOut":
        return cls(id=group.id, code=group.code, description=get_language_description(group, lang))


class Phase(BaseModel):
    id: UUID4


class GroupWithPhaseIdOut(BaseModel):
    id: UUID4
    code: str
    phase: Phase
    description: str

    @classmethod
    def from_instance(cls, group: "GroupModel", *, lang: Lang) -> "GroupWithPhaseIdOut":
        return cls(
            id=group.id,
            code=group.code,
            phase=Phase(id=group.phase_id),
            description=get_language_description(group, lang),
        )


class AllGroupsResponse(BaseModel):
    phases: List[PhaseOut]
    groups: List[GroupWithPhaseIdOut]


class GroupResponse(BaseModel):
    phase: PhaseOut
    group: GroupOut


class GroupsByPhaseCodeResponse(BaseModel):
    phase: PhaseOut
    groups: List[GroupOut]



import sys
from functools import lru_cache

import pendulum

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

from pydantic import HttpUrl, PlainValidator
from pydantic_settings import BaseSettings, SettingsConfigDict

from .rules import Rules

PendulumDateTime = Annotated[pendulum.DateTime, PlainValidator(pendulum.parse)]


class Settings(BaseSettings):
    jwt_secret_key: str
    jwt_expiration_time: int
    competition: str
    lock_datetime: PendulumDateTime
    data_folder: str
    rules: Rules
    official_results_url: HttpUrl

    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="allow")


@lru_cache(maxsize=None)
def get_settings() -> Settings:
    return Settings()



from typing import List

from pydantic import BaseModel

from .binary_bets import BinaryBetOut, BinaryBetWithGroupIdOut
from .group_rank import GroupPositionOut
from .groups import GroupOut, GroupWithPhaseIdOut
from .phases import PhaseOut
from .score_bets import ScoreBetOut, ScoreBetWithGroupIdOut


class AllBetsResponse(BaseModel):
    phases: List[PhaseOut]
    groups: List[GroupWithPhaseIdOut]
    score_bets: List[ScoreBetWithGroupIdOut]
    binary_bets: List[BinaryBetWithGroupIdOut]


class BetsByPhaseCodeResponse(BaseModel):
    phase: PhaseOut
    groups: List[GroupOut]
    score_bets: List[ScoreBetWithGroupIdOut]
    binary_bets: List[BinaryBetWithGroupIdOut]


class BetsByGroupCodeResponse(BaseModel):
    phase: PhaseOut
    group: GroupOut
    score_bets: List[ScoreBetOut]
    binary_bets: List[BinaryBetOut]


class GroupRankResponse(BaseModel):
    phase: PhaseOut
    group: GroupOut
    group_rank: List[GroupPositionOut]



from typing import List, Optional
from uuid import UUID

import pendulum
import strawberry
from sqlalchemy.orm import Session

from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    GroupPositionModel,
    MatchModel,
    PhaseModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)
from yak_server.helpers.bet_locking import is_locked
from yak_server.helpers.group_position import compute_group_rank


@strawberry.type
class Result:
    instance: strawberry.Private[UserModel]

    number_match_guess: int
    number_score_guess: int
    number_qualified_teams_guess: int
    number_first_qualified_guess: int
    number_quarter_final_guess: int
    number_semi_final_guess: int
    number_final_guess: int
    number_winner_guess: int
    points: float

    @classmethod
    def from_instance(cls, instance: UserModel) -> "Result":
        return cls(
            instance=instance,
            number_match_guess=instance.number_match_guess,
            number_score_guess=instance.number_score_guess,
            number_qualified_teams_guess=instance.number_qualified_teams_guess,
            number_first_qualified_guess=instance.number_first_qualified_guess,
            number_quarter_final_guess=instance.number_quarter_final_guess,
            number_semi_final_guess=instance.number_semi_final_guess,
            number_final_guess=instance.number_final_guess,
            number_winner_guess=instance.number_winner_guess,
            points=instance.points,
        )


@strawberry.type
class UserWithoutSensitiveInfo:
    instance: strawberry.Private[UserModel]

    first_name: str
    last_name: str

    result: Result

    @strawberry.field
    def full_name(self) -> str:
        return f"{self.first_name} {self.last_name}"

    @classmethod
    def from_instance(cls, instance: UserModel) -> "UserWithoutSensitiveInfo":
        return cls(
            instance=instance,
            first_name=instance.first_name,
            last_name=instance.last_name,
            result=Result.from_instance(instance),
        )


@strawberry.type
class Flag:
    url: str


@strawberry.type
class Team:
    instance: strawberry.Private[TeamModel]

    id: UUID
    code: str
    description: str
    flag: Flag

    @classmethod
    def from_instance(cls, instance: TeamModel) -> "Team":
        return cls(
            instance=instance,
            id=instance.id,
            code=instance.code,
            description=instance.description_fr,
            flag=Flag(url=instance.flag_url),
        )


@strawberry.type
class TeamWithScore(Team):
    score: Optional[int]

    @classmethod
    def from_instance(cls, instance: TeamModel, *, score: Optional[int]) -> "TeamWithScore":
        return cls(
            instance=instance,
            id=instance.id,
            code=instance.code,
            description=instance.description_fr,
            flag=Flag(url=instance.flag_url),
            score=score,
        )


@strawberry.type
class TeamWithVictory(Team):
    won: Optional[bool]

    @classmethod
    def from_instance(cls, instance: TeamModel, *, won: Optional[bool]) -> "TeamWithVictory":
        return cls(
            instance=instance,
            id=instance.id,
            code=instance.code,
            description=instance.description_fr,
            flag=Flag(url=instance.flag_url),
            won=won,
        )


@strawberry.type
class GroupPosition:
    instance: strawberry.Private[GroupPositionModel]

    team: Team
    played: int
    won: int
    drawn: int
    lost: int
    goals_for: int
    goals_against: int

    @strawberry.field
    def goals_difference(self) -> int:
        return self.goals_for - self.goals_against

    @strawberry.field
    def points(self) -> int:
        return self.won * 3 + self.drawn

    @classmethod
    def from_instance(cls, instance: GroupPositionModel) -> "GroupPosition":
        return cls(
            instance=instance,
            team=Team.from_instance(instance.team),
            played=instance.won + instance.drawn + instance.lost,
            won=instance.won,
            drawn=instance.drawn,
            lost=instance.lost,
            goals_for=instance.goals_for,
            goals_against=instance.goals_against,
        )


def send_group_position(group_rank: List[GroupPositionModel]) -> List[GroupPosition]:
    return sorted(
        [GroupPosition.from_instance(group_position) for group_position in group_rank],
        key=lambda team: (
            team.points(),
            team.goals_difference(),
            team.goals_for,
        ),
        reverse=True,
    )


@strawberry.type
class Group:
    db: strawberry.Private[Session]
    instance: strawberry.Private[GroupModel]
    user: strawberry.Private[UserModel]
    lock_datetime: strawberry.Private[pendulum.DateTime]

    id: UUID
    code: str
    description: str

    @strawberry.field
    def group_rank(self) -> List[GroupPosition]:
        group_rank = self.db.query(GroupPositionModel).filter_by(
            user_id=self.user.id,
            group_id=self.id,
        )

        if not any(group_position.need_recomputation for group_position in group_rank):
            return send_group_position(group_rank)

        score_bets = self.user.score_bets.filter(MatchModel.group_id == self.id).join(
            ScoreBetModel.match,
        )
        group_rank = compute_group_rank(group_rank, score_bets)
        self.db.commit()

        return send_group_position(group_rank)

    @strawberry.field
    def phase(self) -> "Phase":
        return Phase.from_instance(
            self.instance.phase,
            db=self.db,
            user=self.user,
            lock_datetime=self.lock_datetime,
        )

    @strawberry.field
    def score_bets(self) -> List["ScoreBet"]:
        return [
            ScoreBet.from_instance(score_bet, db=self.db, lock_datetime=self.lock_datetime)
            for score_bet in (
                self.db.query(ScoreBetModel)
                .filter_by(user_id=self.user.id)
                .join(ScoreBetModel.match)
                .join(MatchModel.group)
                .filter(
                    MatchModel.group_id == self.instance.id,
                )
                .order_by(GroupModel.index, MatchModel.index)
            )
        ]

    @strawberry.field
    def binary_bets(self) -> List["BinaryBet"]:
        return [
            BinaryBet.from_instance(binary_bet, db=self.db, lock_datetime=self.lock_datetime)
            for binary_bet in (
                self.db.query(BinaryBetModel)
                .filter_by(user_id=self.user.id)
                .join(BinaryBetModel.match)
                .join(MatchModel.group)
                .filter(
                    MatchModel.group_id == self.instance.id,
                )
                .order_by(GroupModel.index, MatchModel.index)
            )
        ]

    @classmethod
    def from_instance(
        cls,
        instance: GroupModel,
        *,
        db: Session,
        user: UserModel,
        lock_datetime: pendulum.DateTime,
    ) -> "Group":
        return cls(
            db=db,
            instance=instance,
            user=user,
            lock_datetime=lock_datetime,
            id=instance.id,
            code=instance.code,
            description=instance.description_fr,
        )


@strawberry.type
class ScoreBet:
    instance: strawberry.Private[ScoreBetModel]

    id: UUID
    locked: bool
    group: Group
    team1: Optional[TeamWithScore]
    team2: Optional[TeamWithScore]

    @classmethod
    def from_instance(
        cls,
        instance: ScoreBetModel,
        *,
        db: "Session",
        lock_datetime: pendulum.DateTime,
    ) -> "ScoreBet":
        return cls(
            instance=instance,
            id=instance.id,
            locked=is_locked(instance.user.name, lock_datetime),
            group=Group.from_instance(
                instance.match.group,
                db=db,
                user=instance.user,
                lock_datetime=lock_datetime,
            ),
            team1=(
                TeamWithScore.from_instance(instance.match.team1, score=instance.score1)
                if instance.match.team1_id is not None
                else None
            ),
            team2=(
                TeamWithScore.from_instance(instance.match.team2, score=instance.score2)
                if instance.match.team2_id is not None
                else None
            ),
        )


@strawberry.type
class BinaryBet:
    instance: strawberry.Private[BinaryBetModel]

    id: UUID
    locked: bool
    group: Group
    team1: Optional[TeamWithVictory]
    team2: Optional[TeamWithVictory]

    @classmethod
    def from_instance(
        cls,
        instance: BinaryBetModel,
        *,
        db: "Session",
        lock_datetime: pendulum.DateTime,
    ) -> "BinaryBet":
        bet_results = instance.bet_from_is_one_won()

        return cls(
            instance=instance,
            id=instance.id,
            locked=is_locked(instance.user.name, lock_datetime),
            group=Group.from_instance(
                instance.match.group,
                db=db,
                user=instance.user,
                lock_datetime=lock_datetime,
            ),
            team1=(
                TeamWithVictory.from_instance(instance.match.team1, won=bet_results[0])
                if instance.match.team1_id is not None
                else None
            ),
            team2=(
                TeamWithVictory.from_instance(instance.match.team2, won=bet_results[1])
                if instance.match.team2_id is not None
                else None
            ),
        )


@strawberry.type
class Phase:
    db: strawberry.Private[Session]
    instance: strawberry.Private[PhaseModel]
    user: strawberry.Private[UserModel]
    lock_datetime: strawberry.Private[pendulum.DateTime]

    id: UUID
    code: str
    description: str

    @strawberry.field
    def groups(self) -> List[Group]:
        return [
            Group.from_instance(group, db=self.db, user=self.user, lock_datetime=self.lock_datetime)
            for group in self.db.query(GroupModel)
            .filter_by(phase_id=self.instance.id)
            .order_by(
                GroupModel.index,
            )
        ]

    @strawberry.field
    def binary_bets(self) -> List["BinaryBet"]:
        return [
            BinaryBet.from_instance(binary_bet, db=self.db, lock_datetime=self.lock_datetime)
            for binary_bet in self.db.query(BinaryBetModel)
            .filter_by(user_id=self.user.id)
            .join(BinaryBetModel.match)
            .join(MatchModel.group)
            .filter(
                GroupModel.phase_id == self.instance.id,
            )
            .order_by(GroupModel.index, MatchModel.index)
        ]

    @strawberry.field
    def score_bets(self) -> List["ScoreBet"]:
        return [
            ScoreBet.from_instance(score_bet, db=self.db, lock_datetime=self.lock_datetime)
            for score_bet in self.db.query(ScoreBetModel)
            .filter_by(user_id=self.user.id)
            .join(ScoreBetModel.match)
            .join(MatchModel.group)
            .filter(
                GroupModel.phase_id == self.instance.id,
            )
            .order_by(GroupModel.index, MatchModel.index)
        ]

    @classmethod
    def from_instance(
        cls,
        instance: PhaseModel,
        *,
        db: Session,
        user: UserModel,
        lock_datetime: pendulum.DateTime,
    ) -> "Phase":
        return cls(
            db=db,
            instance=instance,
            user=user,
            lock_datetime=lock_datetime,
            id=instance.id,
            code=instance.code,
            description=instance.description_fr,
        )


@strawberry.type
class User:
    instance: strawberry.Private[UserModel]
    db: strawberry.Private[Session]
    lock_datetime: strawberry.Private[pendulum.DateTime]

    id: UUID
    pseudo: str
    first_name: str
    last_name: str

    result: Result

    @strawberry.field
    def full_name(self) -> str:
        return f"{self.first_name} {self.last_name}"

    @strawberry.field
    def binary_bets(self) -> List["BinaryBet"]:
        return [
            BinaryBet.from_instance(binary_bet, db=self.db, lock_datetime=self.lock_datetime)
            for binary_bet in self.db.query(BinaryBetModel)
            .filter_by(user_id=self.instance.id)
            .join(BinaryBetModel.match)
            .join(MatchModel.group)
            .order_by(GroupModel.index, MatchModel.index)
        ]

    @strawberry.field
    def score_bets(self) -> List["ScoreBet"]:
        return [
            ScoreBet.from_instance(score_bet, db=self.db, lock_datetime=self.lock_datetime)
            for score_bet in self.db.query(ScoreBetModel)
            .filter_by(user_id=self.instance.id)
            .join(ScoreBetModel.match)
            .join(MatchModel.group)
            .order_by(GroupModel.index, MatchModel.index)
        ]

    @strawberry.field
    def groups(self) -> List["Group"]:
        return [
            Group.from_instance(
                group,
                db=self.db,
                user=self.instance,
                lock_datetime=self.lock_datetime,
            )
            for group in self.db.query(GroupModel).order_by(GroupModel.index)
        ]

    @strawberry.field
    def phases(self) -> List["Phase"]:
        return [
            Phase.from_instance(
                phase,
                db=self.db,
                user=self.instance,
                lock_datetime=self.lock_datetime,
            )
            for phase in self.db.query(PhaseModel).order_by(PhaseModel.index)
        ]

    @classmethod
    def from_instance(
        cls,
        instance: UserModel,
        *,
        db: Session,
        lock_datetime: pendulum.DateTime,
    ) -> "User":
        return cls(
            instance=instance,
            db=db,
            lock_datetime=lock_datetime,
            id=instance.id,
            pseudo=instance.name,
            first_name=instance.first_name,
            last_name=instance.last_name,
            result=Result.from_instance(instance),
        )


@strawberry.type
class UserWithToken(User):
    token: str

    @classmethod
    def from_instance(
        cls,
        instance: UserModel,
        *,
        db: Session,
        lock_datetime: pendulum.DateTime,
        token: str,
    ) -> "UserWithToken":
        return cls(
            instance=instance,
            db=db,
            lock_datetime=lock_datetime,
            id=instance.id,
            pseudo=instance.name,
            first_name=instance.first_name,
            last_name=instance.last_name,
            result=Result.from_instance(instance),
            token=token,
        )



import logging
import sys

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

import pendulum
from fastapi import APIRouter, Depends, status
from pydantic import UUID4
from sqlalchemy.orm import Session

from yak_server.database.models import (
    MatchModel,
    MatchReferenceModel,
    ScoreBetModel,
    UserModel,
)
from yak_server.helpers.authentication import encode_bearer_token
from yak_server.helpers.database import get_db
from yak_server.helpers.group_position import create_group_position
from yak_server.helpers.logging import (
    logged_in_successfully,
    modify_password_successfully,
    signed_up_successfully,
)
from yak_server.helpers.settings import Settings, get_settings
from yak_server.v1.helpers.auth import get_admin_user, get_current_user
from yak_server.v1.helpers.errors import InvalidCredentials, NameAlreadyExists, UserNotFound
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.users import (
    CurrentUserOut,
    LoginIn,
    LoginOut,
    ModifyUserIn,
    SignupIn,
    SignupOut,
)

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/users", tags=["users"])


def signup_user(db: Session, signup_in: SignupIn) -> UserModel:
    # Check existing user in db
    existing_user = db.query(UserModel).filter_by(name=signup_in.name).first()
    if existing_user:
        raise NameAlreadyExists(signup_in.name)

    # Initialize user and integrate in db
    user = UserModel(
        signup_in.name,
        signup_in.first_name,
        signup_in.last_name,
        signup_in.password,
    )
    db.add(user)
    db.flush()

    # Initialize matches and bets and integrate in db
    for match_reference in db.query(MatchReferenceModel).all():
        match = MatchModel(
            team1_id=match_reference.team1_id,
            team2_id=match_reference.team2_id,
            index=match_reference.index,
            group_id=match_reference.group_id,
        )
        db.add(match)
        db.flush()

        db.add(match_reference.bet_type_from_match.value(user_id=user.id, match_id=match.id))
        db.flush()

    # Create group position records
    db.add_all(create_group_position(db.query(ScoreBetModel).filter_by(user_id=user.id)))
    db.commit()

    logger.info(signed_up_successfully(user.name))

    return user


@router.post("/signup", status_code=status.HTTP_201_CREATED)
def signup(
    signup_in: SignupIn,
    db: Annotated[Session, Depends(get_db)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> GenericOut[SignupOut]:
    user = signup_user(db, signup_in)

    return GenericOut(
        result=SignupOut(
            id=user.id,
            name=user.name,
            token=encode_bearer_token(
                sub=user.id,
                expiration_time=pendulum.duration(seconds=settings.jwt_expiration_time),
                secret_key=settings.jwt_secret_key,
            ),
        ),
    )


@router.post("/login", status_code=status.HTTP_201_CREATED)
def login(
    login_in: LoginIn,
    db: Annotated[Session, Depends(get_db)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> GenericOut[LoginOut]:
    user = UserModel.authenticate(db, login_in.name, login_in.password)

    if not user:
        raise InvalidCredentials

    logger.info(logged_in_successfully(user.name))

    return GenericOut(
        result=LoginOut(
            id=user.id,
            name=user.name,
            token=encode_bearer_token(
                sub=user.id,
                expiration_time=pendulum.duration(seconds=settings.jwt_expiration_time),
                secret_key=settings.jwt_secret_key,
            ),
        ),
    )


@router.patch("/{user_id}")
def modify_user(
    user_id: UUID4,
    modify_user_in: ModifyUserIn,
    db: Annotated[Session, Depends(get_db)],
    _: Annotated[UserModel, Depends(get_admin_user)],
) -> GenericOut[CurrentUserOut]:
    user = db.query(UserModel).filter_by(id=user_id).first()
    if not user:
        raise UserNotFound(user_id)

    user.change_password(modify_user_in.password)

    db.commit()

    logger.info(modify_password_successfully(user.name))

    return GenericOut(result=CurrentUserOut.model_validate(user))


@router.get("/current")
def current_user(
    user: Annotated[UserModel, Depends(get_current_user)],
) -> GenericOut[CurrentUserOut]:
    return GenericOut(result=CurrentUserOut.model_validate(user))



from typing import Generator

from sqlalchemy.orm import Session

from yak_server.database import SessionLocal


def get_db() -> Generator[Session, None, None]:
    db_connection = SessionLocal()

    try:
        yield db_connection
    finally:
        db_connection.close()



import typer

from yak_server import create_app

from .database import (
    backup_database,
    compute_score_board,
    create_admin,
    create_database,
    delete_database,
    drop_database,
    initialize_database,
    setup_migration,
)
from .database.sync import synchronize_official_results
from .env import init_env

app = typer.Typer()


def make_db_app() -> typer.Typer:
    db_app = typer.Typer()

    @db_app.command()
    def create() -> None:
        """Create all database tables."""
        create_database()

    @db_app.command()
    def init() -> None:
        """Initialize database."""
        app = create_app()
        initialize_database(app)

    @db_app.command()
    def drop() -> None:
        """Drop all tables."""
        app = create_app()
        drop_database(app)

    @db_app.command()
    def delete() -> None:
        """Delete all records."""
        app = create_app()
        delete_database(app)

    @db_app.command()
    def admin(
        password: str = typer.Option(..., prompt=True, confirmation_prompt=True, hide_input=True),
    ) -> None:
        """Create admin account in database."""
        create_admin(password)

    @db_app.command()
    def backup() -> None:
        """Backup database in a sql file."""
        backup_database()

    @db_app.command()
    def migration() -> None:
        """Help to run database migration scripts."""
        setup_migration()

    @db_app.command()
    def score_board() -> None:
        """Compute score board."""
        compute_score_board()

    @db_app.command()
    def sync() -> None:
        """Synchronize official results and push them to admin with web
        scraping the world cup wikipedia page"""
        synchronize_official_results()

    return db_app


def make_env_typer() -> typer.Typer:
    env_typer = typer.Typer()

    @env_typer.command()
    def init() -> None:
        """Build the env files you need to start the server."""
        init_env()

    return env_typer


app.add_typer(make_db_app(), name="db")
app.add_typer(make_env_typer(), name="env")


if __name__ == "__main__":  # pragma: no cover
    app()



import sys
from typing import List

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated

from fastapi import APIRouter, Depends
from pydantic import UUID4
from sqlalchemy.orm import Session

from yak_server.database.models import PhaseModel, UserModel
from yak_server.helpers.database import get_db
from yak_server.helpers.language import DEFAULT_LANGUAGE, Lang
from yak_server.v1.helpers.auth import get_current_user
from yak_server.v1.helpers.errors import PhaseNotFound
from yak_server.v1.models.generic import GenericOut
from yak_server.v1.models.phases import PhaseOut

router = APIRouter(prefix="/phases", tags=["phases"])


@router.get("/")
def retrieve_all_phases(
    _: Annotated[UserModel, Depends(get_current_user)],
    db: Annotated[Session, Depends(get_db)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[List[PhaseOut]]:
    return GenericOut(
        result=[
            PhaseOut.from_instance(phase, lang=lang)
            for phase in db.query(PhaseModel).order_by(PhaseModel.index)
        ],
    )


@router.get("/{phase_id}")
def retrieve_phase(
    phase_id: UUID4,
    _: Annotated[UserModel, Depends(get_current_user)],
    db: Annotated[Session, Depends(get_db)],
    lang: Lang = DEFAULT_LANGUAGE,
) -> GenericOut[PhaseOut]:
    phase = db.query(PhaseModel).filter_by(id=phase_id).first()

    if not phase:
        raise PhaseNotFound(phase_id)

    return GenericOut(
        result=PhaseOut.from_instance(phase, lang=lang),
    )



from functools import wraps

from jwt import ExpiredSignatureError, PyJWTError
from strawberry.types import Info

from yak_server.database.models import UserModel
from yak_server.helpers.authentication import decode_bearer_token

from .context import YakContext
from .result import ExpiredToken, InvalidToken, UnauthorizedAccessToAdminAPI

NUMBER_ELEMENTS_IN_AUTHORIZATION = 2


def is_authenticated(f):  # noqa: ANN001, ANN201
    @wraps(f)
    def _verify(*args, info: Info[YakContext, None], **kwargs):
        auth_headers = info.context.request.headers.get("Authorization", "").split()
        db = info.context.db
        settings = info.context.settings

        if len(auth_headers) != NUMBER_ELEMENTS_IN_AUTHORIZATION or auth_headers[0] != "Bearer":
            return InvalidToken()

        token = auth_headers[1]
        try:
            data = decode_bearer_token(token, settings.jwt_secret_key)
        except ExpiredSignatureError:
            return ExpiredToken()
        except PyJWTError:
            return InvalidToken()

        user = db.query(UserModel).filter_by(id=data["sub"]).first()
        if not user:
            return InvalidToken()

        info.context.user = user

        return f(*args, info=info, **kwargs)

    return _verify


def is_admin_authenticated(f):  # noqa: ANN001, ANN201
    @wraps(f)
    def _verify(*args, info: Info[YakContext, None], **kwargs):
        user = info.context.user

        if user.name != "admin":
            return UnauthorizedAccessToAdminAPI()

        return f(*args, info=info, **kwargs)

    return _verify



from typing import Any, Dict
from uuid import UUID

import pendulum
from jwt import decode as jwt_decode
from jwt import encode as jwt_encode


def encode_bearer_token(
    sub: UUID,
    expiration_time: pendulum.Duration,
    secret_key: str,
) -> str:
    return jwt_encode(
        {
            "sub": str(sub),
            "iat": pendulum.now("UTC"),
            "exp": pendulum.now("UTC") + expiration_time,
        },
        secret_key,
        algorithm="HS512",
    )


def decode_bearer_token(token: str, secret_key: str) -> Dict[str, Any]:
    return jwt_decode(token, secret_key, algorithms=["HS512"])



import json
import logging
import subprocess  # noqa: S404
from pathlib import Path
from typing import TYPE_CHECKING

import pendulum

from yak_server.database import Base, SessionLocal, engine, mysql_settings
from yak_server.database.models import (
    BinaryBetModel,
    GroupModel,
    GroupPositionModel,
    MatchModel,
    MatchReferenceModel,
    PhaseModel,
    ScoreBetModel,
    TeamModel,
    UserModel,
)
from yak_server.helpers.logging import setup_logging
from yak_server.helpers.rules.compute_points import compute_points as compute_points_func
from yak_server.helpers.settings import get_settings
from yak_server.v1.models.users import SignupIn
from yak_server.v1.routers.users import signup_user

try:
    import alembic
except ImportError:  # pragma: no cover
    alembic = None  # type: ignore[assignment]

if TYPE_CHECKING:
    from fastapi import FastAPI

logger = logging.getLogger(__name__)


class RecordDeletionInProductionError(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to delete records in production using script. DO NOT DO IT.")


class TableDropInProductionError(Exception):
    def __init__(self) -> None:
        super().__init__("Trying to drop database tables in production using script. DO NOT DO IT.")


class BackupError(Exception):
    def __init__(self, description: str) -> None:
        super().__init__(f"Error during backup. {description}")


def create_database() -> None:
    with SessionLocal():
        Base.metadata.create_all(bind=engine)


def create_admin(password: str) -> None:
    with SessionLocal() as db:
        _ = signup_user(
            db,
            SignupIn(name="admin", first_name="admin", last_name="admin", password=password),
        )


def initialize_database(app: "FastAPI") -> None:
    with SessionLocal() as db:
        data_folder = get_settings().data_folder

        phases = json.loads(Path(data_folder, "phases.json").read_text())

        db.add_all(PhaseModel(**phase) for phase in phases)
        db.flush()

        groups = json.loads(Path(data_folder, "groups.json").read_text())

        for group in groups:
            phase = db.query(PhaseModel).filter_by(code=group.pop("phase_code")).first()
            group["phase_id"] = phase.id

        db.add_all(GroupModel(**group) for group in groups)
        db.flush()

        teams = json.loads(Path(data_folder, "teams.json").read_text())

        for team in teams:
            team["flag_url"] = ""

            team_instance = TeamModel(**team)
            db.add(team_instance)
            db.flush()

            team_instance.flag_url = app.url_path_for(
                "retrieve_team_flag_by_id",
                team_id=team_instance.id,
            )
            db.flush()

        matches = json.loads(Path(data_folder, "matches.json").read_text())

        for match in matches:
            team1_code = match.pop("team1_code")
            team2_code = match.pop("team2_code")

            if team1_code is None:
                match["team1_id"] = None
            else:
                team1 = db.query(TeamModel).filter_by(code=team1_code).first()
                match["team1_id"] = team1.id

            if team2_code is None:
                match["team2_id"] = None
            else:
                team2 = db.query(TeamModel).filter_by(code=team2_code).first()
                match["team2_id"] = team2.id

            group = db.query(GroupModel).filter_by(code=match.pop("group_code")).first()
            match["group_id"] = group.id

        db.add_all(MatchReferenceModel(**match) for match in matches)
        db.flush()

        db.commit()


def backup_database() -> None:
    setup_logging(debug=False)

    result = subprocess.run(
        [  # noqa: S603, S607
            "mysqldump",
            mysql_settings.db,
            "-u",
            mysql_settings.user_name,
            "-P",
            str(mysql_settings.port),
            "--protocol=tcp",
            f"--password={mysql_settings.password}",
        ],
        capture_output=True,
        encoding="utf-8",
        check=False,
    )

    backup_datetime = pendulum.now("UTC")

    if result.returncode:
        error_message = (
            f"Something went wrong when backup on {backup_datetime}: "
            f"{result.stderr.replace(mysql_settings.password, '********')}"
        )

        logger.error(error_message)

        raise BackupError(error_message)

    backup_location = Path(__file__).parents[1] / "backup_files"
    backup_location.mkdir(exist_ok=True)

    file_path = (
        backup_location / f"yak_toto_backup_{backup_datetime.format('YYYYMMDD[T]HHmmssZZ')}.sql"
    )

    file_path.write_text(result.stdout)
    logger.info(f"Backup done on {backup_datetime}")


def delete_database(app: "FastAPI") -> None:
    if not app.debug:
        raise RecordDeletionInProductionError

    with SessionLocal() as db:
        db.query(GroupPositionModel).delete()
        db.query(ScoreBetModel).delete()
        db.query(BinaryBetModel).delete()
        db.query(UserModel).delete()
        db.query(MatchReferenceModel).delete()
        db.query(MatchModel).delete()
        db.query(GroupModel).delete()
        db.query(PhaseModel).delete()
        db.query(TeamModel).delete()
        db.commit()


def drop_database(app: "FastAPI") -> None:
    if not app.debug:
        raise TableDropInProductionError

    with SessionLocal():
        Base.metadata.drop_all(bind=engine)


def setup_migration() -> None:
    alembic_ini_path = Path(__file__).parents[3] / "alembic.ini"

    print(
        "To be able to run the database migration scripts, you need to run the following command:",
    )
    print(f"export ALEMBIC_CONFIG='{alembic_ini_path.resolve()}'")
    print()
    print(
        "Follow this link for more informations: "
        "https://alembic.sqlalchemy.org/en/latest/tutorial.html#editing-the-ini-file",
    )

    if alembic is None:
        print()
        print("To enable migration using alembic, please run: pip install yak-server[db_migration]")


def compute_score_board() -> None:
    with SessionLocal() as db:
        admin = db.query(UserModel).filter_by(name="admin").first()

        rule_config = get_settings().rules.compute_points

        compute_points_func(db, admin, rule_config)
